/*
 * This file is part of the SPiC Project SPiChess  (https://gitlab.cs.fau.de/i4/spic/chess).
 * Copyright (c) 2017 - 2020 by Bernhard Heinloth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "board.h"

#include <assert.h>

#include "chess.h"
#include "ai.h"
#include "openings.h"

#define BOARD_SELECTED_FROM board[selected.fromY][selected.fromX]
#define BOARD_SELECTED_TO board[selected.toY][selected.toX]

static const int8_t board_default[8][8] = {
	{  4,  2,  3,  5,  6,  3,  2,  4 },
	{  1,  1,  1,  1,  1,  1,  1,  1 },
	{  0,  0,  0,  0,  0,  0,  0,  0 },
	{  0,  0,  0,  0,  0,  0,  0,  0 },
	{  0,  0,  0,  0,  0,  0,  0,  0 },
	{  0,  0,  0,  0,  0,  0,  0,  0 },
	{ -1, -1, -1, -1, -1, -1, -1, -1 },
	{ -4, -2, -3, -5, -6, -3, -2, -4 }
};

int8_t board[8][8];
static uint16_t counter[3];
chess_move selected;
static chess_move last;

#define ROCHADE_ROOK_DISABLE(X,P)  rochade |= 1 << ((((P) >> 7) & 1) * 2 + ((X) & 1))
#define ROCHADE_ROOK_POSSIBLE(X,P)  (((rochade >> ((((P) >> 7) & 1) * 2 + ((X) & 1))) & 1) == 0)
static uint8_t rochade;

void board_reset() {
	openings_reset();
	counter[0] = 0;
	counter[1] = 0;
	counter[2] = 0;
	for (uint8_t y = 0; y < 8; y++)
		for (uint8_t x = 0; x < 8; x++)
			board[y][x] = board_default[y][x];
	last = selected = SELECTED_INVALID;
	rochade = 0;
}

bool board_isDraw() {
	return counter[1] - counter[0] > 100 || counter[1] - counter[2] > 100;
}

bool board_hasKing(int8_t player) {
	for (uint8_t y = 0; y < 8; y++)
		for (uint8_t x = 0; x < 8; x++)
			if (board[y][x] == KING * player)
				return true;
	return false;
}


uint16_t board_moveCounter() {
	return counter[1];
}


#define CAPTURE_MARK(Y,X) if (VALID_FIELD(Y,X)) captureBoard |= ((uint64_t)1 << ((Y) * 8 + (X)))
#define CAPTURE_CHECK(Y,X) (((captureBoard >> ((Y) * 8 + (X))) & 1) == 1)
static uint64_t board_capture(int8_t player) {
	uint64_t captureBoard = 0llu;
	for (int y = 0; y < 8; y++)
		for (int x = 0; x < 8; x++) {
			uint8_t fig = board[y][x] * player * (-1);
			switch (fig) {
				case PAWN:
					CAPTURE_MARK(y - player, x + 1);
					CAPTURE_MARK(y - player, x - 1);
					break;
				case KNIGHT:
					CAPTURE_MARK(y + 2, x + 1);
					CAPTURE_MARK(y + 2, x - 1);
					CAPTURE_MARK(y - 2, x + 1);
					CAPTURE_MARK(y - 2, x - 1);
					CAPTURE_MARK(y + 1, x + 2);
					CAPTURE_MARK(y + 1, x - 2);
					CAPTURE_MARK(y - 1, x + 2);
					CAPTURE_MARK(y - 1, x - 2);
					break;
				case KING:
					CAPTURE_MARK(y, x + 1);
					CAPTURE_MARK(y + 1, x + 1);
					CAPTURE_MARK(y + 1, x);
					CAPTURE_MARK(y + 1, x - 1);
					CAPTURE_MARK(y, x - 1);
					CAPTURE_MARK(y - 1, x + 1);
					CAPTURE_MARK(y - 1, x);
					CAPTURE_MARK(y - 1, x - 1);
					break;
				case QUEEN:
				case ROOK:
					for (int8_t i = x + 1; i < 8 && (captureBoard |= (uint64_t) 1 << ((y) * 8 + (i))) != 0 && board[y][i] == NONE; i++) ;
					for (int8_t i = x - 1; i >= 0 && (captureBoard |= (uint64_t) 1 << ((y) * 8 + (i))) != 0 && board[y][i] == NONE; i--) ;
					for (int8_t i = y + 1; i < 8 && (captureBoard |= (uint64_t) 1 << ((i) * 8 + (x))) != 0 && board[i][x] == NONE; i++) ;
					for (int8_t i = y - 1; i >= 0 && (captureBoard |= (uint64_t) 1 << ((i) * 8 + (x))) != 0 && board[i][x] == NONE; i--) ;
					if (fig == ROOK)
						break;
				case BISHOP:
					// bishop or queen
					for (int8_t i = 1; y + i < 8 && x + i < 8 && (captureBoard |= (uint64_t) 1 << ((y + i) * 8 + (x + i))) != 0 && board[y + i][x + i] == NONE; i++) ;
					for (int8_t i = 1; y - i >= 0 && x - i >= 0 && (captureBoard |= (uint64_t) 1 << ((y - i) * 8 + (x - i))) != 0 && board[y - i][x - i] == NONE; i++) ;
					for (int8_t i = 1; y + i < 8 && x - i >= 0 && (captureBoard |= (uint64_t) 1 << ((y + i) * 8 + (x - i))) != 0 && board[y + i][x - i] == NONE; i++) ;
					for (int8_t i = 1; y - i >= 0 && x + i < 8 && (captureBoard |= (uint64_t) 1 << ((y - i) * 8 + (x + i))) != 0 && board[y - i][x + i] == NONE; i++) ;
					break;
			}
		}
	return captureBoard;
}

bool board_isCheck(int8_t player){
	for (uint8_t y = 0; y < 8; y++)
		for (uint8_t x = 0; x < 8; x++)
			if (board[y][x] == KING * player){
				uint64_t captureBoard = board_capture(player);
				return CAPTURE_CHECK(y,x);
			}
	return false;
}

__attribute__ ((hot))
bool board_canMove(uint8_t fromY, uint8_t fromX, uint8_t toY, uint8_t toX, int8_t player) {
	assert(VALID_FIELD(fromY, fromX) && VALID_FIELD(toY, toX) >= 0 && fromX < 8 && toY >= 0 && toY < 8 && toX >= 0 && toX < 8 && player * player == 1);
	if (fromY == toY && fromX == toX)
		return false;

	int8_t from = board[fromY][fromX];
	int8_t to = board[toY][toX];
	FIGURES fig = from * player;
	if (from == 0 || from * player <= 0 || to * player > 0)
		return false;
	else if (fig == PAWN) {
		if (fromX == toX && to == 0 && (toY - fromY == player || (toY - fromY == 2 * player && board[player + fromY][fromX] == 0 && fromY == (player > 0 ? 1 : 6))))
			return true;
		// capture (inc. en passant)
		else
			return ((fromX - toX) * (fromX - toX) == 1 && toY - fromY == player) && (to * player < 0 || (board[last.toY][last.toX] == PAWN * (-1) * player && (last.toY - last.fromY) * (last.toY - last.fromY) == 4 && last.toY == fromY && last.toX == last.fromX && last.toX == toX));
	} else if (fig == QUEEN || fig == ROOK || fig == BISHOP) {
		// rook - and queen
		if (fig != BISHOP) {
			if (fromX == toX) {
				for (uint8_t i = (fromY < toY ? fromY : toY) + 1; i < (fromY > toY ? fromY : toY); i++)
					if (board[i][fromX] != NONE)
						return false;
				return true;
			} else if (fromY == toY) {
				for (uint8_t i = (fromX < toX ? fromX : toX) + 1; i < (fromX > toX ? fromX : toX); i++)
					if (board[fromY][i] != NONE)
						return false;
				return true;
			}
		}
		// bishop - and queen
		if (fig != ROOK) {
			if ((fromY - toY) * (fromY - toY) != (fromX - toX) * (fromX - toX))
				return false;
			for (int8_t i = 1; i < (fromY > toY ? fromY - toY : toY - fromY); i++)
				if (board[fromY + i * (toY - fromY > 0 ? 1 : -1)][fromX + i * (toX - fromX > 0 ? 1 : -1)] != NONE)
					return false;
			return true;
		}
		// nothing found? wrong move.
		return false;
	} else if (fig == KNIGHT)
		return (((fromY - toY == 2 || fromY - toY == -2) && (fromX - toX == 1 || fromX - toX == -1)) || ((fromY - toY == 1 || fromY - toY == -1) && (fromX - toX == 2 || fromX - toX == -2)));
	else if (fig == KING) {
		uint64_t captureBoard = board_capture(player);
		// king must not be in check after move
		if (CAPTURE_CHECK(toY, toX))
			return false;
		// Check castling
		if (fromX == 4 && fromY == (player < 0 ? 7 : 0) && toY == fromY && !CAPTURE_CHECK(fromY, fromX) && ((toX == 2 && board[fromY][0] == ROOK * player && board[fromY][1] == NONE && board[fromY][2] == NONE && board[fromY][3] == NONE && ROCHADE_ROOK_POSSIBLE(0, player) && !CAPTURE_CHECK(fromY, 3)) || (toX == 6 && board[fromY][7] == ROOK * player && board[fromY][6] == NONE && board[fromY][5] == NONE && ROCHADE_ROOK_POSSIBLE(7, player) && !CAPTURE_CHECK(fromY, 5))))	//   ) ))
			return true;
		else
			return (((fromX - toX) * (fromX - toX)) <= 1 && ((fromY - toY) * (fromY - toY)) <= 1);
	}
	// invalid figure
	else
		return error(20);
}

uint8_t board_move(int8_t player) {
	if (board_canMove(selected.fromY, selected.fromX, selected.toY, selected.toX, player)) {
		uint8_t m = MOVE_VALID;

		// En Passant
		if (BOARD_SELECTED_FROM == player && (selected.fromX - selected.toX) * (selected.fromX - selected.toX) == 1 && selected.toY - selected.fromY == player && (last.toY - last.fromY) * (last.toY - last.fromY) == 4 && board[last.toY][last.toX] == -1 * player && last.toY == selected.fromY && last.toX == last.fromX && last.toX == selected.toX){
			board[last.toY][last.toX] = 0;
			m |= MOVE_EN_PASSANT;
		}
		// castling disable figures
		else if (BOARD_SELECTED_FROM == ROOK * player)
			ROCHADE_ROOK_DISABLE(selected.fromX, player);
		else if (BOARD_SELECTED_FROM == KING * player && selected.fromX == 4) {
			ROCHADE_ROOK_DISABLE(0, player);
			ROCHADE_ROOK_DISABLE(7, player);
			// Perform castling
			if (selected.toX == 2) {
				assert(board[selected.toY][0] == ROOK * player);
				board[selected.toY][3] = board[selected.toY][0];
				board[selected.toY][0] = NONE;
				m |= MOVE_CASTELING_QUEENSIDE;
			} else if (selected.toX == 6) {
				assert(board[selected.toY][7] == ROOK * player);
				board[selected.toY][5] = board[selected.toY][7];
				board[selected.toY][7] = NONE;
				m |= MOVE_CASTELING_KINGSIDE;
			}
		}
		// Fifty moves rule
		if (BOARD_SELECTED_TO != NONE || (m & MOVE_EN_PASSANT) != 0){
			counter[player + 1] = counter[1];
			m |=  MOVE_CAPTURE;
		}

		// Pawn to Queen
		if (BOARD_SELECTED_FROM == player && selected.toY == (player > 0 ? 7 : 0)){
			BOARD_SELECTED_TO = QUEEN * player;
			m |= MOVE_PROMOTION;
		}
		// Move
		else
			BOARD_SELECTED_TO = BOARD_SELECTED_FROM;

		BOARD_SELECTED_FROM = 0;
		// Copy move
		last = selected;
		counter[1]++;
		return m;
	} else
		return MOVE_INVALID;
}
