/*
 * This file is part of the SPiC Project SPiChess  (https://gitlab.cs.fau.de/i4/spic/chess).
 * Copyright (c) 2017 - 2020 by Bernhard Heinloth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdlib.h>
#include <stdbool.h>

#include "7seg.h"
#include "adc.h"
#include "button.h"
#include "led.h"

#include "ai.h"
#include "board.h"
#include "openings.h"
#include "main_display.h"


#define LED_WON(PLAYER) ((PLAYER) <= 0 ? (1 << RED0) : (1 << RED1))
#define LED_TURN(PLAYER) ((PLAYER) <= 0 ? (1 << YELLOW0) : (1 << YELLOW1))
#define LED_TURN_FROM(PLAYER) (((PLAYER) <= 0 ? (1 << YELLOW0) : (1 << YELLOW1)) | ((PLAYER) <= 0 ? (1 << GREEN0) : (1 << GREEN1)))
#define LED_TURN_TO(PLAYER) (((PLAYER) <= 0 ? (1 << YELLOW0) : (1 << YELLOW1)) | ((PLAYER) <= 0 ? (1 << BLUE0) : (1 << BLUE1)))
#define LED_SELECT(PLAYER) ((PLAYER) <= 0 ? ((1 << RED0) | (1 << YELLOW0) | (1 << GREEN0) | (1 << BLUE0)) : ((1 << RED1) | (1 << YELLOW1) | (1 << GREEN1) | (1 << BLUE1)))
#define LED_DRAW ((1 << RED0) | (1 << RED1))

#define INVALID ((uint8_t)-1)

// ai strength - 0 = human player
#define DEFAULT_AI_STRENGTH 4
static uint8_t ai_strength[2] = { DEFAULT_AI_STRENGTH, 0 };
#define AI_STRENGTH(PLAYER) ai_strength[(PLAYER) <= 0 ? 0 : 1]


static char sb_7seg_text[3] = { ' ', ' ', '\0' };
// Helper to print chars
static int8_t sb_7seg_showChars(char first, char second){
	if (first != '\0')
		sb_7seg_text[0] = first;
	if (second != '\0')
		sb_7seg_text[1] = second;
	return sb_7seg_showString(sb_7seg_text);
}

// Helper to receive Buttons which status has changed from released to pressed since the last call
static bool sb_button_wasPressed(BUTTON b) {
	static uint8_t previousState[2] = {UNKNOWN, UNKNOWN};
	uint8_t currentState = sb_button_getState(b);
	bool r = currentState == PRESSED && previousState[b] == RELEASED;
	previousState[b] = currentState;
	return r;
}

// Helper to read a value using the potientiometer (and print the current value)
static int8_t inputField(int8_t y, int8_t x) {
	int8_t prev = -1;
	while (true) {
		int8_t value = 7 - (sb_adc_read(POTI) * (uint16_t) 8 / 1024);
		sb_7seg_text[y < 0] = (y < 0 ? '1' : 'a') + value;
		sb_7seg_showString(sb_7seg_text);
		// Highlight on display
		display_field(y < 0 ? value : y, x < 0 ? value : x, true);
		// undo previous highlight (unless it is the from position, in which case we are at the to-ones)
		if (prev >= 0 && prev != value && !(selected.fromY == y && selected.fromX == x))
			display_field(y < 0 ? prev : y, x < 0 ? prev : x, false);
		prev = value;

		// Finish on button press
		if (sb_button_wasPressed(BUTTON0)){
			display_field(y < 0 ? value : y, x < 0 ? value : x, false);
			return value;
		} else if (sb_button_wasPressed(BUTTON1)){
			sb_7seg_text[y < 0] = ' ';
			sb_7seg_showString(sb_7seg_text);
			display_field(y < 0 ? value : y, x < 0 ? value : x, false);
			return -1;
		} else
			_delay_ms(20);
	};
}

// Helper to set up the player
static bool getPlayerSetup(const int8_t player) {
	display_playerMenu(player);
	if (player < 0)
		return true;
	sb_led_setMask(LED_SELECT(player));
	sb_7seg_showChars('P', '2' - player);
	_delay_ms(1000);
	while (true){
		int8_t value = 5 - (sb_adc_read(POTI) * (uint16_t) 6 / 1024);
		sb_7seg_showChars(value == 0 ? 'h' : 'L', value == 0 ? 'u' : (value + '0'));
		display_playerMenuSelect(value);
		if (sb_button_wasPressed(BUTTON0)){
			if (!getPlayerSetup(player - 1))
				break;
			AI_STRENGTH(player) = value;
			return true;
		} else if (sb_button_wasPressed(BUTTON1)){
			break;
		} else
			_delay_ms(20);
	};
	sb_7seg_showChars(' ', ' ');
	return false;
}


int8_t error(uint8_t e) {
	char first = (e / 16) + (e / 16 < 10 ? 48 : 55);
	char second = (e % 16) + (e % 16 < 10 ? 48 : 55);
	sb_7seg_showChars(first, second);
	sb_led_setMask(0xff);
	char str[] = { 'S', 'Y', 'S', 'T', 'E', 'M', 'F', 'E', 'H', 'L', 'E', 'R', ' ', first, second, '\0' };
	display_status(str);
	while (true) ;
	return 1;
}


static bool move(int8_t player) {
	if (AI_STRENGTH(player) == 0){
		char movestr[] = { 'F', 'i', 'g', 'u', 'r', ' ', '_', '_', ' ', 'z', 'u', '?', ' ',  ' ', ' ', '\0' };
		// Player
		while (1) {
			// Reset
			selected = SELECTED_INVALID;
			display_board();
			sb_led_setMask(LED_TURN(player));
			sb_7seg_showChars('H', 'u');
			display_statusFromFlash(PSTR("Du bist am Zug."));
			_delay_ms(1000);
			while (1) {
				sb_led_setMask(LED_TURN_FROM(player));
				sb_7seg_showChars(' ', ' ');
				if ((selected.fromX = inputField(player <= 0 ? 7 : 0, -1)) == INVALID)
					break;
				else if ((selected.fromY = inputField(-1, selected.fromX)) != INVALID) {
					if (board[selected.fromY][selected.fromX] == 0) {
						sb_7seg_showChars('-', '-');
						display_statusFromFlash(PSTR("Feld ist leer! "));
						_delay_ms(1000);
					} else if (board[selected.fromY][selected.fromX] * player < 0) {
						sb_7seg_showChars('P', 'C');
						display_statusFromFlash(PSTR("Falsche Farbe! "));
						_delay_ms(1000);
					} else {
						display_field(selected.fromY, selected.fromX, true);
						_delay_ms(1000);
						sb_led_setMask(LED_TURN(player));
						sb_7seg_showChars('t', 'o');

						movestr[6] = 'a' + selected.fromX;
						movestr[7] = '1' + selected.fromY;
						display_status(movestr);

						_delay_ms(1000);
						sb_led_setMask(LED_TURN_TO(player));
						sb_7seg_showChars(' ', ' ');
						while ((selected.toX = inputField(selected.fromY, -1)) != INVALID) {
							if ((selected.toY = inputField(-1, selected.toX)) != INVALID) {
								display_selected();

								int8_t m = board_move(player);
								// Invalid move
								if (m == MOVE_INVALID) {
									display_field(selected.toY, selected.toX, false);
									sb_7seg_showChars('I', 'n');
									display_statusFromFlash(PSTR("Nicht erlaubt! "));
									_delay_ms(1000);
									break;
								} else {
									// valid move
									sb_7seg_showChars('O', 'k');

									movestr[11] = ' ';
									movestr[12] = 'a' + selected.toX;
									movestr[13] = '1' + selected.toY;
									display_status(movestr);

									_delay_ms(1000);
									display_move(player, m);
									return true;
								}
							}
						}
						break;
					}
				}
			}
		}
	} else {
		// AI
		sb_led_setMask(LED_TURN(player));
		display_board();
		display_statusFromFlash(PSTR("PC denkt nach. "));
		sb_7seg_disable();
		ai_move(player, AI_STRENGTH(player));
		if (SELECTED_IS_INVALID)
			return false;
		else
			display_selected();

		char movestr[] = { 'P', 'C', ':', ' ', 'a' + selected.fromX, selected.fromY + '1', ' ', 'n', 'a', 'c', 'h', ' ', 'a' + selected.toX, selected.toY + '1', '!', '\0' };
		display_status(movestr);

		for (int8_t i = 0;; i = (i + 1) % 4) {
			switch (i) {
				case 1:
					sb_led_setMask(LED_TURN_FROM(player));
					sb_7seg_showChars('a' + selected.fromX, selected.fromY + '1');
					break;
				case 2:
					sb_led_setMask(LED_TURN(player));
					sb_7seg_showChars('t', 'o');
					break;
				case 3:
					sb_led_setMask(LED_TURN_TO(player));
					sb_7seg_showChars('a' + selected.toX, selected.toY + '1');
					break;
				default:
					sb_led_setMask(LED_TURN(player));
					sb_7seg_showChars('P', 'C');
					break;
			}
			for (int8_t j = 0; j < 50; j++) {
				if (sb_button_wasPressed(BUTTON0)) {
					int8_t m = board_move(player);
					if (m == MOVE_INVALID)
						error(0xfe);
					else
						display_move(player, m);
					return true;
				} else
					_delay_ms(20);
			}
		}
	}
}

const __flash char title[] = " SPiChess   ";
const __flash char gameOver[] = " Game over  ";
void main(void) {
#ifndef NDEBUG
	sb_7seg_showChars('C', 'K');
	openings_validate();
#endif
	sei();

	do {
		// Play intro
		display_intro();
		uint8_t leds = 0xFF;
		for (uint8_t i = 0; i < sizeof(title) - 2; i++) {
			sb_led_setMask(leds);
			sb_7seg_showChars(title[i], title[i + 1]);
			_delay_ms(400);
			leds <<= 1;
		}
	} while (!getPlayerSetup(1));

	// Main Game Loop
	board_reset();
	display_board();
	int8_t player = -1;
	while (true) {
		player *= -1;
		if (board_isDraw()) {
			sb_led_setMask(LED_DRAW);
			display_statusFromFlash(PSTR("Unentschieden! "));
			break;
		} else if (!board_hasKing(player) || !move(player)) {
			sb_led_setMask(LED_WON(player * (-1)));
			display_statusFromFlash(AI_STRENGTH(player) == 0 ? PSTR("Der PC gewinnt.") : PSTR(" Du gewinnst!  "));
			break;
		}
	}

	// Game over animation
	display_board();
	for (uint8_t i = 0;; i = (i + 1) % (sizeof(gameOver) - 2)) {
		sb_7seg_showChars(gameOver[i], gameOver[i + 1]);
		_delay_ms(400);
	}
}
